package fr.ippon.direct.cache;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

/**
 * @author IPPON_NANTES
 * 
 */
public class NativeMemoryCache {

	/**
	 * Création de la map pour stocker la clé ainsi que l'objet en mémoire
	 * native sous forme de ByteBuffer
	 */
	private Map<Object, ByteBuffer> data = new HashMap<Object, ByteBuffer>();

	/**
	 * Ajoute un objet à la Map
	 * 
	 * @param key
	 *            clé unique de l'objet
	 * @param object
	 *            : objet à ajouter à la map
	 * @throws IOException
	 */
	public void put(Object key, Serializable object) throws IOException {
		byte[] bytes = serialize(object);
		// on alloue de la mémoire pour le stockage de l'objet
		ByteBuffer buf = ByteBuffer.allocateDirect(bytes.length);
		buf.put(bytes);
		buf.flip();
		data.put(key, buf);
	}

	/**
	 * 
	 * @param key
	 *            clé unique de l'objet
	 * @return objet correspondant à la clef
	 * @throws ClassNotFoundException
	 * @throws IOException
	 */
	public Object get(Object key) throws ClassNotFoundException, IOException {
		ByteBuffer buf = data.get(key).duplicate();
		byte[] bytes = new byte[buf.remaining()];
		buf.get(bytes);
		return deserialize(bytes);
	}

	/**
	 * Permet de lister le contenu de la Hashmap
	 * 
	 * @throws ClassNotFoundException
	 * @throws IOException
	 */
	public void print() throws ClassNotFoundException, IOException {
		for (Iterator<Map.Entry<Object, ByteBuffer>> it = data.entrySet()
				.iterator(); it.hasNext();) {
			Map.Entry<Object, ByteBuffer> e = (Entry<Object, ByteBuffer>) it
					.next();
			ByteBuffer buf = (ByteBuffer) e.getValue();
			byte[] bytes = new byte[buf.remaining()];
			buf.get(bytes);

			System.out.println(deserialize(bytes));
		}

	}

	/**
	 * Méthode de libération de la mémoire allouée par allocateDirect()
	 * 
	 * @throws IOException
	 * @throws NoSuchMethodException
	 * @throws SecurityException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws InvocationTargetException
	 */
	public void freeBuffer() throws IOException, NoSuchMethodException,
			SecurityException, IllegalAccessException,
			IllegalArgumentException, InvocationTargetException {
		for (Iterator<Map.Entry<Object, ByteBuffer>> it = data.entrySet()
				.iterator(); it.hasNext();) {
			Map.Entry<Object, ByteBuffer> e = (Entry<Object, ByteBuffer>) it
					.next();

			ByteBuffer buffer = (ByteBuffer) e.getValue();
			Method getCleanerMethod = buffer.getClass().getMethod("cleaner",
					new Class[0]);
			getCleanerMethod.setAccessible(true);
			sun.misc.Cleaner cleaner = (sun.misc.Cleaner) getCleanerMethod
					.invoke(buffer, new Object[0]);
			cleaner.clean();
		}
	}

	/**
	 * 
	 * @param obj
	 *            : Objet à sérialiser
	 * @return objet sérialisé
	 * @throws IOException
	 */
	private byte[] serialize(Object obj) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(obj);
		return baos.toByteArray();
	}

	/**
	 * 
	 * @param bytes
	 *            Objet sous forme sérialisée
	 * @return Objet
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private Object deserialize(byte[] bytes) throws IOException,
			ClassNotFoundException {
		ByteArrayInputStream bis = new ByteArrayInputStream(bytes);
		ObjectInputStream ois = new ObjectInputStream(bis);
		Serializable obj = (Serializable) ois.readObject();
		ois.close();
		return obj;

	}

}
